from onegov.reservation.collection import ResourceCollection
from onegov.reservation.core import LibresIntegration
from onegov.reservation.models import CustomAllocation as Allocation
from onegov.reservation.models import CustomReservation as Reservation
from onegov.reservation.models import Resource

__all__ = [
    'Allocation',
    'LibresIntegration',
    'Reservation',
    'Resource',
    'ResourceCollection',
]



import pytest

from libres.context.registry import create_default_registry
from libres.db.models import ORMBase
from onegov.reservation import LibresIntegration
from uuid import uuid4


@pytest.yield_fixture(scope="function")
def libres_context(session_manager):

    session_manager.bases.append(ORMBase)
    session_manager.set_current_schema('test_' + uuid4().hex)

    registry = create_default_registry()

    yield LibresIntegration.libres_context_from_session_manager(
        registry, session_manager
    )



import pytest

from datetime import date, datetime
from onegov.reservation import ResourceCollection


def test_resource_collection(libres_context):

    collection = ResourceCollection(libres_context)
    assert collection.query().count() == 0

    resource = collection.add('Executive Lounge', 'Europe/Zurich')
    assert resource.name == 'executive-lounge'
    assert resource.timezone == 'Europe/Zurich'
    assert resource.get_scheduler(libres_context).resource == resource.id

    assert collection.query().count() == 1
    assert collection.by_id(resource.id)
    assert collection.by_name('executive-lounge')

    collection.delete(collection.by_id(resource.id))

    assert collection.query().count() == 0


def test_resource_save_delete(libres_context):
    collection = ResourceCollection(libres_context)

    resource = collection.add('Executive Lounge', 'Europe/Zurich')
    assert resource.name == 'executive-lounge'
    assert resource.timezone == 'Europe/Zurich'
    scheduler = resource.get_scheduler(libres_context)

    dates = (datetime(2015, 8, 5, 12), datetime(2015, 8, 5, 18))

    scheduler.allocate(dates)
    scheduler.reserve('info@example.org', dates)

    with pytest.raises(AssertionError):
        collection.delete(resource)

    collection.delete(resource, including_reservations=True)

    assert collection.query().count() == 0
    assert scheduler.managed_reservations().count() == 0
    assert scheduler.managed_reserved_slots().count() == 0
    assert scheduler.managed_allocations().count() == 0


def test_resource_highlight_allocations(libres_context):
    collection = ResourceCollection(libres_context)
    resource = collection.add('Executive Lounge', 'Europe/Zurich')

    assert resource.date is None
    assert resource.highlights_min is None
    assert resource.highlights_max is None

    scheduler = resource.get_scheduler(libres_context)
    dates = (datetime(2015, 8, 5, 12), datetime(2015, 8, 5, 18))
    allocations = scheduler.allocate(dates)

    resource.highlight_allocations(allocations)

    assert resource.date == date(2015, 8, 5)
    assert resource.highlights_min == allocations[0].id
    assert resource.highlights_min == allocations[-1].id


def test_resource_form_definition(libres_context):
    collection = ResourceCollection(libres_context)

    resource = collection.add(
        title='Executive Lounge',
        timezone='Europe/Zurich',
        definition='Mail *= @@@'
    )
    assert resource.form_class().mail is not None

    resource.definition = None
    assert resource.form_class is None



from datetime import datetime
from onegov.reservation import ResourceCollection


def test_per_hour_pricing(libres_context):
    collection = ResourceCollection(libres_context)

    resource = collection.add('Executive Lounge', 'Europe/Zurich')
    resource.currency = 'CHF'
    resource.pricing_method = 'per_hour'
    resource.price_per_hour = 10

    scheduler = resource.get_scheduler(libres_context)
    dates = (datetime(2017, 6, 7, 12), datetime(2017, 6, 7, 18))
    scheduler.allocate(dates)

    token = scheduler.reserve('info@example.org', dates)
    reservation = scheduler.reservations_by_token(token).one()

    assert reservation.price().amount == 60
    assert reservation.price().currency == 'CHF'
    assert reservation.price(resource).amount == 60
    assert reservation.price(resource).currency == 'CHF'


def test_multiple_allocations(libres_context):
    collection = ResourceCollection(libres_context)

    resource = collection.add('Executive Lounge', 'Europe/Zurich')
    resource.currency = 'CHF'
    resource.pricing_method = 'per_hour'
    resource.price_per_hour = 10

    scheduler = resource.get_scheduler(libres_context)
    dates = (
        (datetime(2017, 6, 7, 12), datetime(2017, 6, 7, 18)),
        (datetime(2017, 6, 8, 12), datetime(2017, 6, 8, 18)),
    )
    scheduler.allocate(dates)

    token = scheduler.reserve('info@example.org', dates)
    reservations = scheduler.reservations_by_token(token).all()

    assert reservations[0].price().amount == 60
    assert reservations[0].price().currency == 'CHF'
    assert reservations[1].price().amount == 60
    assert reservations[1].price().currency == 'CHF'
    assert reservations[0].price(resource).amount == 60
    assert reservations[0].price(resource).currency == 'CHF'
    assert reservations[1].price(resource).amount == 60
    assert reservations[1].price(resource).currency == 'CHF'


def test_per_reservation_pricing(libres_context):
    collection = ResourceCollection(libres_context)

    resource = collection.add('Executive Lounge', 'Europe/Zurich')
    resource.currency = 'CHF'
    resource.pricing_method = 'per_item'
    resource.price_per_item = 10

    scheduler = resource.get_scheduler(libres_context)
    dates = (datetime(2017, 6, 7, 12), datetime(2017, 6, 7, 18))
    scheduler.allocate(dates, quota=2)

    token = scheduler.reserve('info@example.org', dates, quota=2)
    reservation = scheduler.reservations_by_token(token).one()

    assert reservation.price().amount == 20
    assert reservation.price().currency == 'CHF'
    assert reservation.price(resource).amount == 20
    assert reservation.price(resource).currency == 'CHF'



import pytest
import transaction

from datetime import datetime
from freezegun import freeze_time
from onegov.reservation.models import Resource
from pytz import utc
from uuid import uuid4


def test_libres_context_fixture(session_manager, libres_context):
    assert session_manager is libres_context.get_service('session_provider')


def test_resource_scheduler(libres_context):
    resource = Resource(id=uuid4())

    with pytest.raises(AssertionError):
        resource.get_scheduler(libres_context)

    resource.timezone = 'Europe/Zurich'

    scheduler = resource.get_scheduler(libres_context)
    scheduler.managed_allocations().count() == 0

    assert scheduler.resource == resource.id


def test_scheduler_boundaries(libres_context):
    resource = Resource(id=uuid4())
    resource.timezone = 'Europe/Amsterdam'

    scheduler = resource.get_scheduler(libres_context)
    scheduler.allocate((
        datetime(2015, 6, 11, 8),
        datetime(2015, 6, 11, 18)
    ))

    assert scheduler.managed_allocations().count() == 1

    new_resource = Resource()
    new_resource.id = resource.id
    new_resource.timezone = 'Europe/Amsterdam'

    scheduler = new_resource.get_scheduler(libres_context)

    assert scheduler.managed_allocations().count() == 1

    new_resource = Resource(id=uuid4())
    new_resource.timezone = 'Europe/Amsterdam'
    new_resource.libres_context = libres_context

    scheduler = new_resource.get_scheduler(libres_context)

    assert scheduler.managed_allocations().count() == 0


def test_delete_cascade(session_manager, libres_context):
    resource = Resource(id=uuid4())
    resource.timezone = 'Europe/Zurich'
    resource.name = 'test'
    resource.title = 'Test'

    session = session_manager.session()
    session.add(resource)

    scheduler = resource.get_scheduler(libres_context)
    scheduler.allocate((
        datetime(2015, 6, 11, 8),
        datetime(2015, 6, 11, 18)
    ))

    transaction.commit()
    session.expire_all()

    resource = session.query(Resource).one()
    scheduler = resource.get_scheduler(libres_context)
    assert scheduler.managed_allocations().count() == 1

    session.delete(resource)

    assert scheduler.managed_allocations().count() == 0


def test_invalid_deadlines():
    resource = Resource()

    with pytest.raises(ValueError):
        resource.deadline = 'Foobar'

    with pytest.raises(ValueError):
        resource.deadline = '3h'

    with pytest.raises(ValueError):
        resource.deadline = ('3', 'd')

    with pytest.raises(ValueError):
        resource.deadline = (0, 'd')

    with pytest.raises(ValueError):
        resource.deadline = (3, 'm')

    resource.deadline = ''
    assert resource.deadline is None

    resource.deadline = None
    resource.deadline = (5, 'd')
    resource.deadline = (24, 'h')


def test_deadline():
    resource = Resource(timezone='UTC')

    # by default, no deadline is active
    with freeze_time("2018-11-28"):
        assert not resource.is_past_deadline(
            datetime(2018, 11, 27, tzinfo=utc))
        assert not resource.is_past_deadline(
            datetime(2018, 11, 28, tzinfo=utc))
        assert not resource.is_past_deadline(
            datetime(2018, 11, 29, tzinfo=utc))

    allocation = datetime(2018, 11, 28, 13, tzinfo=utc)

    resource.deadline = (1, 'h')

    # if it's 11:59, we can reserve for 13:00
    with freeze_time("2018-11-28 11:59"):
        assert not resource.is_past_deadline(allocation)

    # if it's 12:00, we can no longer reserve for 13:00
    with freeze_time("2018-11-28 12:00"):
        assert resource.is_past_deadline(allocation)

    # if it's 13:00, we can no longer reserve for 13:00
    with freeze_time("2018-11-28 13:00"):
        assert resource.is_past_deadline(allocation)

    resource.deadline = (1, 'd')

    # the day before at 23:59, we can still reserve
    with freeze_time("2018-11-27 23:59"):
        assert not resource.is_past_deadline(allocation)

    # the same day at 00:00 we can no longer reserve
    with freeze_time("2018-11-28 00:00"):
        assert resource.is_past_deadline(allocation)

    # any time thereafter is the same
    with freeze_time("2018-11-28 13:00"):
        assert resource.is_past_deadline(allocation)

    with freeze_time("2018-11-29 13:00"):
        assert resource.is_past_deadline(allocation)



from onegov.reservation.models.resource import Resource
from onegov.reservation.models.custom_allocation import CustomAllocation
from onegov.reservation.models.custom_reservation import CustomReservation

__all__ = ('CustomAllocation', 'CustomReservation', 'Resource')



""" Contains upgrade tasks that are executed when the application is being
upgraded on the server. See :class:`onegov.core.upgrade.upgrade_task`.

"""
from libres.db.models import Allocation, Reservation
from onegov.core.upgrade import upgrade_task
from onegov.reservation import LibresIntegration
from onegov.reservation import Resource
from sqlalchemy import Column, Text


def run_upgrades(context):
    """ onegov.reservation is a bit special because it defines its tables
    through its own declarative base. This is due to libres requireing its own
    base.

    As a consequence, not all applications loaded in the current process have
    all the tables for all the modules (which is usually the case for all
    onegov modules using the default onegov.core.orm.Base class).

    This means we can only run the upgrades if Libres is integrated with
    the current app.

    """
    return isinstance(context.app, LibresIntegration)


@upgrade_task('Add form definition field')
def add_form_definition_field(context):

    if run_upgrades(context):
        context.operations.add_column(
            'resources', Column('definition', Text, nullable=True)
        )


@upgrade_task('Add resource group field')
def add_resource_group_field(context):

    if run_upgrades(context):
        context.operations.add_column(
            'resources', Column('group', Text, nullable=True)
        )


@upgrade_task('Add reservations/allocations type field')
def add_reservations_allocations_type_field(context):

    if run_upgrades(context):
        context.operations.add_column(
            'reservations', Column('type', Text, nullable=True)
        )
        context.operations.add_column(
            'allocations', Column('type', Text, nullable=True)
        )


@upgrade_task('Make reservations/allocations payable')
def make_reservations_allocations_payable(context):

    if run_upgrades(context):
        for reservation in context.session.query(Reservation):
            reservation.type = 'custom'

        for allocation in context.session.query(Allocation):
            allocation.type = 'custom'


@upgrade_task('Set defaults on existing resources')
def set_defaults_on_existing_reservation_resourcd_objects(context):

    if run_upgrades(context):
        for resource in context.session.query(Resource):
            resource.payment_method = 'manual'
            resource.pricing_method = 'free'
            resource.price_per_hour = 0
            resource.price_per_item = 0
            resource.currency = 'CHF'



from datetime import timedelta
from onegov.core.cache import lru_cache
from libres import new_scheduler
from libres.db.models import Allocation
from libres.db.models.base import ORMBase
from onegov.core.orm import ModelBase
from onegov.core.orm.mixins import content_property
from onegov.core.orm.mixins import ContentMixin, TimestampMixin
from onegov.core.orm.types import UUID
from onegov.form import parse_form
from onegov.pay import Price, process_payment
from sedate import align_date_to_day, utcnow
from sqlalchemy import Column, Text
from sqlalchemy.orm import relationship
from uuid import uuid4


@lru_cache(maxsize=1)
def extra_scheduler_arguments():
    from onegov.reservation.models import CustomReservation
    from onegov.reservation.models import CustomAllocation

    return dict(
        allocation_cls=CustomAllocation,
        reservation_cls=CustomReservation
    )


class Resource(ORMBase, ModelBase, ContentMixin, TimestampMixin):
    """ A resource holds a single calendar with allocations and reservations.

    Note that this resource is not defined on the onegov.core declarative base.
    Instead it is defined using the libres base. This means we can't join
    data outside the libres models.

    This should however not be a problem as this onegov module is self
    contained and does not link to other onegov modules, except for core.

    If we ever want to link to other models (say link a reservation to a user),
    then we have to switch to a unified base. Ideally we would find a way
    to merge these bases somehow.

    Also note that we *do* use the ModelBase class as a mixin to at least share
    the same methods as all the usual onegov.core.orm models.

    """

    __tablename__ = 'resources'

    #: the unique id
    id = Column(UUID, primary_key=True, default=uuid4)

    #: a nice id for the url, readable by humans
    name = Column(Text, primary_key=False, unique=True)

    #: the title of the resource
    title = Column(Text, primary_key=False, nullable=False)

    #: the timezone this resource resides in
    timezone = Column(Text, nullable=False)

    #: the custom form definition used when creating a reservation
    definition = Column(Text, nullable=True)

    #: the group to which this resource belongs to (may be any kind of string)
    group = Column(Text, nullable=True)

    #: the type of the resource, this can be used to create custom polymorphic
    #: subclasses. See `<http://docs.sqlalchemy.org/en/improve_toc/
    #: orm/extensions/declarative/inheritance.html>`_.
    type = Column(Text, nullable=True)

    #: the payment method
    payment_method = content_property()

    #: the currency of the price to pay
    currency = content_property()

    #: the pricing method to use
    pricing_method = content_property()

    #: the reservations cost a given amount per hour
    price_per_hour = content_property()

    #: the reservations cost a given amount per unit (allocations * quota)
    price_per_item = content_property('price_per_reservation')

    #: reservation deadline (e.g. None, (5, 'd'), (24, 'h'))
    deadline = content_property()

    __mapper_args__ = {
        "polymorphic_on": 'type'
    }

    allocations = relationship(
        Allocation,
        cascade="all, delete-orphan",
        primaryjoin='Resource.id == Allocation.resource',
        foreign_keys='Allocation.resource'
    )

    #: the date to jump to in the view (if not None) -> not in the db!
    date = None

    #: a list of allocations ids to highlight in the view (if not None)
    highlights_min = None
    highlights_max = None

    #: the view to open in the calendar (fullCalendar view name)
    view = 'month'

    @deadline.setter
    def set_deadline(self, value):
        value = value or None

        if value:
            if len(value) != 2:
                raise ValueError("Deadline is not a tuple with two elements")

            if type(value[0]) != int:
                raise ValueError("Deadline value is not an int")

            if value[0] < 1:
                raise ValueError("Deadline value is smaller than 1")

            if value[1] not in ('d', 'h'):
                raise ValueError("Deadline unit must be 'd' or 'h'")

        self.content['deadline'] = value

    def highlight_allocations(self, allocations):
        """ The allocation to jump to in the view. """

        # we can assume that allocation ids are created in a continuous
        # number line. It's not necessarily guaranteed, but since it *is*
        # only a highlighting feature we can check the highlights more
        # effiecently if we follow this assumption.
        highlights = [a.id for a in allocations]

        self.highlights_min = min(highlights)
        self.highlights_max = max(highlights)

        self.date = allocations[0].start.date()

    def get_scheduler(self, libres_context):
        assert self.id, "the id needs to be set"
        assert self.timezone, "the timezone needs to be set"

        return new_scheduler(
            libres_context, self.id, self.timezone,
            **extra_scheduler_arguments()
        )

    @property
    def scheduler(self):
        assert hasattr(self, 'libres_context'), "not bound to libres context"
        return self.get_scheduler(self.libres_context)

    def bind_to_libres_context(self, libres_context):
        self.libres_context = libres_context

    @property
    def form_class(self):
        """ Parses the form definition and returns a form class. """

        if not self.definition:
            return None

        return parse_form(self.definition)

    def price_of_reservation(self, token, extra=None):
        reservations = self.scheduler.reservations_by_token(token)

        prices = (r.price(self) for r in reservations)
        prices = (p for p in prices if p)

        total = sum(prices, Price.zero())

        if extra and total:
            total += extra
        elif extra:
            total = extra

        return total

    def process_payment(self, price, provider=None, payment_token=None):
        """ Processes the payment for the given reservation token. """

        if price and price.amount > 0:
            return process_payment(
                self.payment_method, price, provider, payment_token)

        return True

    def is_past_deadline(self, date):
        if not self.deadline:
            return False

        if not date.tzinfo:
            raise RuntimeError(f"The given date has no timezone: {date}")

        if not self.timezone:
            raise RuntimeError(f"No timezone set on the resource")

        n, unit = self.deadline

        # hours result in a simple offset
        def deadline_using_h():
            return date - timedelta(hours=n)

        # days require that we align the date to the beginning of the date
        def deadline_using_d():
            return align_date_to_day(date, self.timezone, 'down')\
                - timedelta(days=(n - 1))

        deadline = locals()[f'deadline_using_{unit}']()
        return deadline <= utcnow()



from libres.db.models import Allocation
from onegov.core.orm import ModelBase
from onegov.reservation.models.resource import Resource
from sqlalchemy.orm import object_session


class CustomAllocation(Allocation, ModelBase):
    __mapper_args__ = {'polymorphic_identity': 'custom'}

    @property
    def resource_obj(self):
        return object_session(self).query(Resource)\
            .filter_by(id=self.resource).one()



from datetime import timedelta
from libres.db.models import Reservation
from onegov.core.orm import ModelBase
from onegov.pay import Payable, Price
from onegov.reservation.models.resource import Resource
from sqlalchemy.orm import object_session


class CustomReservation(Reservation, ModelBase, Payable):
    __mapper_args__ = {'polymorphic_identity': 'custom'}

    @property
    def resource_obj(self):
        return object_session(self).query(Resource)\
            .filter_by(id=self.resource).one()

    @property
    def payable_reference(self):
        return f'{self.resource.hex}/{self.email}x{self.quota}'

    def price(self, resource=None):
        """ Returns the price of the reservation.

        Even though one token may point to multiple reservations the price
        is bound to the reservation record.

        The price per token is calculcated by combining all the prices.

        """

        resource = resource or self.resource_obj

        if resource.pricing_method not in ('per_hour', 'per_item'):
            return None

        # technically we could have multiple allocations per reservation
        # but in practice we don't use that feature. Each reservation
        # links to exactly one allocation.
        #
        # As a result we can take a substantial shortcut here and calculate
        # the price on the reservation itself instead of loading all
        # allocations.

        if resource.pricing_method == 'per_hour':
            duration = self.end + timedelta(microseconds=1) - self.start
            hours = duration.total_seconds() // 3600

            return Price(hours * resource.price_per_hour, resource.currency)

        if resource.pricing_method == 'per_item':
            count = self.quota

            return Price(count * resource.price_per_item, resource.currency)

        raise NotImplementedError



import morepath

from datetime import datetime
from libres import new_scheduler
from onegov.core.framework import Framework
from onegov.core.utils import scan_morepath_modules
from onegov.reservation import LibresIntegration, ResourceCollection
from sqlalchemy import Column, Integer
from sqlalchemy.ext.declarative import declarative_base
from uuid import uuid4
from webtest import TestApp as Client
from onegov.core.orm import Base as CoreBase


def test_setup_database(postgres_dsn, redis_url):
    Base = declarative_base()

    class App(Framework, LibresIntegration):
        pass

    class Document(Base):
        __tablename__ = 'documents'
        id = Column(Integer, primary_key=True)

    @App.path(path='/')
    class Root(object):
        pass

    @App.json(model=Root)
    def get_root(self, request):
        return []

    # this is required for the transactions to actually work, usually this
    # would be onegov.server's job
    scan_morepath_modules(App)
    morepath.commit(App)

    app = App()
    app.configure_application(
        dsn=postgres_dsn,
        base=CoreBase,
        redis_url=redis_url
    )
    app.session_manager.bases.append(Base)
    app.namespace = 'libres'
    app.set_application_id('libres/foo')

    c = Client(app)
    c.get('/')

    tables = app.session().execute(
        "SELECT table_name FROM information_schema.tables "
        "WHERE table_schema = 'public'"
    )

    assert not tables.fetchall()

    tables = app.session().execute(
        "SELECT table_name FROM information_schema.tables "
        "WHERE table_schema = 'libres-foo'"
    )

    tables = set(r[0] for r in tables.fetchall())

    assert 'documents' in tables
    assert 'payments_for_reservations_payment' in tables
    assert 'resources' in tables
    assert 'allocations' in tables
    assert 'reserved_slots' in tables
    assert 'reservations' in tables

    app.session_manager.dispose()


def test_libres_context(postgres_dsn):
    Base = declarative_base()

    class App(Framework, LibresIntegration):
        pass

    app = App()
    app.configure_application(dsn=postgres_dsn, base=CoreBase)
    app.session_manager.bases.append(Base)
    app.namespace = 'libres'
    app.set_application_id('libres/foo')
    app.session_manager.set_current_schema('libres-foo')

    tables = app.session().execute(
        "SELECT table_name FROM information_schema.tables "
        "WHERE table_schema = 'libres-foo'"
    )

    tables = set(r[0] for r in tables.fetchall())

    assert 'payments_for_reservations_payment' in tables
    assert 'resources' in tables
    assert 'allocations' in tables
    assert 'reserved_slots' in tables
    assert 'reservations' in tables

    scheduler = new_scheduler(app.libres_context, uuid4(), 'Europe/Zurich')
    assert scheduler.managed_allocations().count() == 0

    scheduler.allocate((datetime(2015, 7, 30, 11), datetime(2015, 7, 30, 12)))
    assert scheduler.managed_allocations().count() == 1

    assert app.session_manager is app.libres_context.get_service(
        'session_provider')


def test_transaction_integration(postgres_dsn, redis_url):
    Base = declarative_base()

    class App(Framework, LibresIntegration):
        pass

    class Document(Base):
        __tablename__ = 'documents'
        id = Column(Integer, primary_key=True)

    @App.path(path='/')
    class Root(object):
        pass

    @App.json(model=Root)
    def handle_root(self, request):
        collection = ResourceCollection(request.app.libres_context)

        resource = collection.add('Test', 'Europe/Zurich')
        scheduler = resource.get_scheduler(request.app.libres_context)
        scheduler.allocate(
            (datetime(2015, 7, 30, 11), datetime(2015, 7, 30, 12))
        )

        # this will fail and then abort everything
        request.session.add(Document(id=1))
        request.session.add(Document(id=1))

    # this is required for the transactions to actually work, usually this
    # would be onegov.server's job
    scan_morepath_modules(App)
    morepath.commit(App)

    app = App()
    app.configure_application(
        dsn=postgres_dsn,
        base=CoreBase,
        redis_url=redis_url
    )
    app.session_manager.bases.append(Base)
    app.namespace = 'libres'
    app.set_application_id('libres/foo')

    c = Client(app)
    try:
        c.get('/', expect_errors=True)
    except Exception:
        pass

    collection = ResourceCollection(app.libres_context)
    assert collection.query().count() == 0



from onegov.core.utils import normalize_for_url
from onegov.reservation.models import Resource
from uuid import uuid4


any_type = object()


class ResourceCollection(object):
    """ Manages a list of resources.

    """
    def __init__(self, libres_context):
        assert hasattr(libres_context, 'get_service'), """
            The ResourceCollection expected the libres_contex, not the session.
        """

        self.libres_context = libres_context
        self.session = libres_context.get_service('session_provider').session()

    def query(self):
        return self.session.query(Resource)

    def add(self, title, timezone, type=None, name=None, meta={}, content={},
            definition=None, group=None):

        # look up the right class depending on the type (we need to do
        # this a bit akwardly here, because Resource does not use the
        # ModelBase as declarative base)
        resource = Resource.get_polymorphic_class(type, Resource)()
        resource.id == uuid4()
        resource.name = name or normalize_for_url(title)
        resource.title = title
        resource.timezone = timezone
        resource.meta = meta
        resource.content = content
        resource.definition = definition
        resource.group = group

        self.session.add(resource)
        self.session.flush()

        return self.bind(resource)

    def bind(self, resource):
        if resource:
            resource.bind_to_libres_context(self.libres_context)

        return resource

    def by_id(self, id, ensure_type=any_type):
        query = self.query().filter(Resource.id == id)

        if ensure_type is not any_type:
            query = query.filter(Resource.type == type)

        return self.bind(query.first())

    def by_name(self, name, ensure_type=any_type):
        query = self.query().filter(Resource.name == name)

        if ensure_type is not any_type:
            query = query.filter(Resource.type == type)

        return self.bind(query.first())

    def by_allocation(self, allocation):
        return self.by_id(allocation.resource)

    def by_reservation(self, reservation):
        return self.by_id(reservation.resource)

    def delete(self, resource, including_reservations=False):
        scheduler = resource.get_scheduler(self.libres_context)

        if not including_reservations:
            assert not scheduler.managed_reserved_slots().first()
            assert not scheduler.managed_reservations().first()

            scheduler.managed_allocations().delete('fetch')
        else:
            scheduler.extinguish_managed_records()

        self.session.delete(resource)
        self.session.flush()



from libres.context.registry import create_default_registry
from libres.db.models import ORMBase
from onegov.reservation.collection import ResourceCollection
from uuid import UUID


class LibresIntegration(object):
    """ Provides libres integration for
    :class:`onegov.core.framework.Framework` based applications.

    The application must be connected to a database

    Usage::

        from onegov.core import Framework

        class MyApp(Framework, LibresIntegration):
            pass

    """

    def configure_libres(self, **cfg):
        """ Configures the libres integration and leaves two properties on
        the class:

        :libres_context:
            The libres context configured for the current application.

        :libres_registrye:
            The libres registry bound to the current application.

        With those two a scheduler can easily be created::

            from libres import new_scheduler
            scheduler = new_scheduler(
                app.libres_context, 'test', 'Europe/Zurich'
            )

        """

        assert self.session_manager.bases, "Must be run after configure_dsn"
        self.session_manager.bases.append(ORMBase)

        self.libres_registry = create_default_registry()
        self.libres_context = self.libres_context_from_session_manager(
            self.libres_registry,
            self.session_manager
        )

    @staticmethod
    def libres_context_from_session_manager(registry, session_manager):
        if registry.is_existing_context('onegov.reservation'):
            return registry.get_context('onegov.reservation')

        context = registry.register_context('onegov.reservation')
        context.set_service('session_provider', lambda ctx: session_manager)

        # onegov.reservation uses uuids for the resources, so we don't need to
        # generate anything, we can just reuse the id (which is passed as the
        # name)
        def uuid_generator(name):
            assert isinstance(name, UUID)
            return name

        context.set_service('uuid_generator', lambda ctx: uuid_generator)

        return context

    @property
    def libres_resources(self):
        return ResourceCollection(self.libres_context)
